/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package javalda;

import java.util.*;
import java.util.Map.*;

/**
 *
 * @author Alfan Farizki Wicaksono
 * Institut Teknologi Bandung
 * Bandung, Indonesia.
 * 
 * Toy Implementation of LDA using Gibbs Sampling
 */
public class LDASampler {
    
    private int     _num_topic;
    private double  _alpha;
    private double  _beta;
    private int     _MAXITER = 100;
    
    /*the number of word in document m assigned by topic z*/
    private int[][] num_mz;
    
    /*the number of word w assigned by topic z in the collection*/
    private int[][] num_zw;
    
    /*the number of word in document m*/
    private int[]   num_m;
    
    /*the number of word assigned with topic z in the collection*/
    private int[]   num_z;
    
    /*Result after gibbs sampling : P(w|z) or PHI*/
    private double[][]  phi_w_z;
    
    private Corpus  crp;
    
    public LDASampler(int num_topic, double alpha, double beta, String vocabFile, String collectionFile) {
        this._num_topic = num_topic;
        this._alpha = alpha;
        this._beta = beta;
        
        this.crp = new Corpus(vocabFile, collectionFile);
        
        int num_docs = crp.getNumDocs();
        int vocab_size = crp.getNumVocab();
        
        /*all elements are ZEROed by default*/
        this.num_mz = new int[num_docs][num_topic];
        this.num_zw = new int[num_topic][vocab_size];
        this.num_m  = new int[num_docs];
        this.num_z  = new int[num_topic];
        
        this.phi_w_z = new double[vocab_size][num_topic];
    }
    
    private void _initialize() {
        
        /*for each document in the collection*/
        for (int m = 0; m < crp.getNumDocs(); m++) {
            Corpus.document d_m = crp.docs.get(m);
            
            /*for each word-topic position in document m*/
            for (Entry w_z: d_m.word_topic.entrySet()) {
                
                /*word in current position*/
                int w = (Integer)w_z.getKey();
                
                /* choose an arbitrary topic as first topic for current word*/
                /* z is between 0 - _num_topic-1 */
                int z = Util.generateTopicRandom(_num_topic);

                this.num_mz[m][z] += 1;
                this.num_m[m] += 1;
                this.num_zw[z][w] += 1;
                this.num_z[z] += 1;
                
                /*assign the choosen topic to current position*/
                w_z.setValue(z);
            }
        }
    }
    
    /*P(zi|Z-i,w)*/
    /*Conditional distribution (vector of size num_topics)*/
    private Map<Integer, Double> probability_z(int m, int w) {
        
        Map<Integer, Double> p_z = new HashMap<Integer, Double>();
        
        int vocab_size = crp.getNumVocab();
        double cumm = 0d;
        
        /*give value P(z|.) for each z*/
        for (int z = 0; z < _num_topic; z++) {
            double LEFT  = (this.num_zw[z][w] + _beta) / (this.num_z[z] + _beta * vocab_size);
            double RIGHT = (this.num_mz[m][z] + _alpha) / (this.num_m[m] + _alpha * _num_topic);
            
            double v = LEFT * RIGHT;
            cumm += v;
            p_z.put(z, v);
        }
        
        //normalizing p_z
        for (Entry z: p_z.entrySet()) {
            double currv = (Double)z.getValue();
            z.setValue(currv/cumm);
        }
        
        return p_z;
    }
    
    public void runGibbsSampling() {
        
        /*initializing, assign random topic to each word FIRST*/
        this._initialize();
        
        /*run iteration to revise the topic assignment for each word*/
        for (int iter = 0; iter < _MAXITER; iter++) {
            
            System.out.println("iteration-"+iter+"...");
            
            /*for each document in the collection*/
            for (int m = 0; m < crp.getNumDocs(); m++) {
                Corpus.document d_m = crp.docs.get(m);
            
                /*for each word-topic position in document m*/
                for (Entry w_z: d_m.word_topic.entrySet()) {

                    /*word in current position*/
                    int w = (Integer)w_z.getKey();
                    
                    /*cancel the assignment of topic of current word*/
                    int z = (Integer)w_z.getValue();
                    
                    this.num_mz[m][z] -= 1;
                    this.num_m[m] -= 1;
                    this.num_zw[z][w] -= 1;
                    this.num_z[z] -= 1;
                    
                    /*multinomial sampling : P(zi|Z-i,w)*/
                    Map<Integer, Double> p_z = probability_z(m, w);
                    z = Util.generateSample(p_z); /*sampling topic from P(zi|Z-i,w)*/

                    /*assign sampled-topic to current word*/
                    w_z.setValue(z);
                    
                    this.num_mz[m][z] += 1;
                    this.num_m[m] += 1;
                    this.num_zw[z][w] += 1;
                    this.num_z[z] += 1;
                }
            }
            
            System.out.println("Likelihood = " + loglikelihood());
        }
        /*BURN-IN !!*/
        
        /*computePHI after burn-in period*/
        System.out.println("P(w|z):");
        this.computePHI();
    }
    
    private double loglikelihood() {
        float likelihood = 0.0f;
        int numvocab = crp.getNumVocab();
        int numdocs  = crp.getNumDocs();
        
        /*num_zw + beta*/
        double[][] num_zw_plus_beta = new double[_num_topic][numvocab];
        for (int z = 0; z < _num_topic; z++) {
            for (int w = 0; w < numvocab; w++) {
                num_zw_plus_beta[z][w] = num_zw[z][w] + _beta;
            }
        }
        
        /*num_mz + alpha*/
        double[][] num_mz_plus_alpha = new double[numdocs][_num_topic];
        for (int m = 0; m < crp.getNumDocs(); m++) {
            for (int z = 0; z < _num_topic; z++) {
                num_mz_plus_alpha[m][z] = num_mz[m][z] + _alpha;
            }
        }
        
        /*NOW ! compute likelihood*/
        for (int z = 0; z < _num_topic; z++) {
            likelihood += Util.logDeltaVector(num_zw_plus_beta[z]);
            likelihood -= Util.logDeltaScalar(_num_topic, _beta);
            
        }
        
        for (int m = 0; m < crp.getNumDocs(); m++) {
            likelihood += Util.logDeltaVector(num_mz_plus_alpha[m]);
            likelihood -= Util.logDeltaScalar(numdocs, _alpha);
        }
        
        return likelihood;
    }
    
    /*PHI = P(w|z)*/
    public void computePHI() {
                
        for (int z = 0; z < _num_topic; z++) {
            
            /*compute num_z + beta*|V|*/
            double denominator = num_z[z] + _beta*crp.getNumVocab();
            
            /*assign each word in P(w|z = z)*/
            for (int w = 0; w < crp.getNumVocab(); w++) {
                phi_w_z[w][z] = (num_zw[z][w] + _beta) / denominator;
                
                System.out.println("P("+w+"|"+z+") = "+phi_w_z[w][z]);
            }
        }
    }
}
